@@ -8,8 +8,6 @@
  *
  * We currently detect the filesystems listed below in the struct
  * "imagetype images" (in the order they are listed).
- *
- * MINIX, ext3 and Reiserfs bits are currently untested.
  */
 
 #include <sys/types.h>
@@ -31,6 +29,9 @@
 #include "xfs_sb.h"
 #include "luks_fs.h"
 #include "lvm2_sb.h"
+#include "iso9660_sb.h"
+#include "squashfs_fs.h"
+#include "gfs2_fs.h"
 
 /*
  * Slightly cleaned up version of jfs_superblock to
@@ -43,6 +44,7 @@
  * Use a cleaned up version.
  */
 #include "reiserfs_fs.h"
+#include "reiser4_fs.h"
 
 #include "fstype.h"
 
@@ -106,6 +108,25 @@ static int minix_image(const void *buf, unsigned long long *bytes)
 	return 0;
 }
 
+static int ext4_image(const void *buf, unsigned long long *bytes)
+{
+	const struct ext3_super_block *sb =
+		(const struct ext3_super_block *)buf;
+
+	if (sb->s_magic == __cpu_to_le16(EXT2_SUPER_MAGIC)
+		&& (sb->s_feature_incompat
+		& __cpu_to_le32(EXT3_FEATURE_INCOMPAT_EXTENTS)
+		|| sb->s_feature_incompat
+		& __cpu_to_le32(EXT4_FEATURE_INCOMPAT_64BIT)
+		|| sb->s_feature_incompat
+		& __cpu_to_le32(EXT4_FEATURE_INCOMPAT_MMP))) {
+		*bytes = (unsigned long long)__le32_to_cpu(sb->s_blocks_count)
+		    << (10 + __le32_to_cpu(sb->s_log_block_size));
+		return 1;
+	}
+	return 0;
+}
+
 static int ext3_image(const void *buf, unsigned long long *bytes)
 {
 	const struct ext3_super_block *sb =
@@ -115,7 +136,7 @@ static int ext3_image(const void *buf, unsigned long long *bytes)
 	    sb->
 	    s_feature_compat & __cpu_to_le32(EXT3_FEATURE_COMPAT_HAS_JOURNAL)) {
 		*bytes = (unsigned long long)__le32_to_cpu(sb->s_blocks_count)
-		    << (10 + sb->s_log_block_size);
+		    << (10 + __le32_to_cpu(sb->s_log_block_size));
 		return 1;
 	}
 	return 0;
@@ -128,7 +149,7 @@ static int ext2_image(const void *buf, unsigned long long *bytes)
 
 	if (sb->s_magic == __cpu_to_le16(EXT2_SUPER_MAGIC)) {
 		*bytes = (unsigned long long)__le32_to_cpu(sb->s_blocks_count)
-		    << (10 + sb->s_log_block_size);
+		    << (10 + __le32_to_cpu(sb->s_log_block_size));
 		return 1;
 	}
 	return 0;
@@ -152,6 +173,20 @@ static int reiserfs_image(const void *buf, unsigned long long *bytes)
 	return 0;
 }
 
+static int reiser4_image(const void *buf, unsigned long long *bytes)
+{
+	const struct reiser4_master_sb *sb =
+		(const struct reiser4_master_sb *)buf;
+
+	if (memcmp(sb->ms_magic, REISER4_SUPER_MAGIC_STRING,
+		sizeof(REISER4_SUPER_MAGIC_STRING) - 1) == 0) {
+		*bytes = (unsigned long long) __le32_to_cpu(sb->ms_format) *
+			__le32_to_cpu(sb->ms_blksize);
+		return 1;
+	}
+	return 0;
+}
+
 static int xfs_image(const void *buf, unsigned long long *bytes)
 {
 	const struct xfs_sb *sb = (const struct xfs_sb *)buf;
@@ -201,6 +236,20 @@ static int swap_image(const void *buf, unsigned long long *blocks)
 	return 0;
 }
 
+static int suspend_image(const void *buf, unsigned long long *blocks)
+{
+	const struct swap_super_block *ssb =
+	    (const struct swap_super_block *)buf;
+
+	if (!memcmp(ssb->magic, SUSP_MAGIC_1, SUSP_MAGIC_L) ||
+	    !memcmp(ssb->magic, SUSP_MAGIC_2, SUSP_MAGIC_L) ||
+	    !memcmp(ssb->magic, SUSP_MAGIC_U, SUSP_MAGIC_L)) {
+		*blocks = 0;
+		return 1;
+	}
+	return 0;
+}
+
 static int lvm2_image(const void *buf, unsigned long long *blocks)
 {
 	const struct lvm2_super_block *lsb;
@@ -221,6 +270,48 @@ static int lvm2_image(const void *buf, unsigned long long *blocks)
 	return 0;
 }
 
+static int iso_image(const void *buf, unsigned long long *blocks)
+{
+	const struct iso_volume_descriptor *isovd =
+	    (const struct iso_volume_descriptor *)buf;
+	const struct iso_hs_volume_descriptor *isohsvd =
+	    (const struct iso_hs_volume_descriptor *)buf;
+
+	if (!memcmp(isovd->id, ISO_MAGIC, ISO_MAGIC_L) ||
+	    !memcmp(isohsvd->id, ISO_HS_MAGIC, ISO_HS_MAGIC_L)) {
+		*blocks = 0;
+		return 1;
+	}
+	return 0;
+}
+
+static int squashfs_image(const void *buf, unsigned long long *blocks)
+{
+	const struct squashfs_super_block *sb =
+		(const struct squashfs_super_block *)buf;
+
+	if (sb->s_magic == SQUASHFS_MAGIC
+	    || sb->s_magic == SQUASHFS_MAGIC_SWAP) {
+		*blocks = (unsigned long long) sb->bytes_used;
+		return 1;
+	}
+	return 0;
+}
+
+static int gfs2_image(const void *buf, unsigned long long *bytes)
+{
+	const struct gfs2_sb *sb =
+		(const struct gfs2_sb *)buf;
+
+	if (__be32_to_cpu(sb->sb_header.mh_magic) == GFS2_MAGIC
+		&& (__be32_to_cpu(sb->sb_fs_format) == GFS2_FORMAT_FS
+		|| __be32_to_cpu(sb->sb_fs_format) == GFS2_FORMAT_MULTI)) {
+		*bytes = 0; /* cpu_to_be32(sb->sb_bsize) * ?; */
+		return 1;
+	}
+	return 0;
+}
+
 struct imagetype {
 	off_t block;
 	const char name[12];
@@ -244,16 +335,22 @@ static struct imagetype images[] = {
 	{0, "cramfs", cramfs_image},
 	{0, "romfs", romfs_image},
 	{0, "xfs", xfs_image},
+	{0, "squashfs", squashfs_image},
+	{1, "ext4dev", ext4_image},
 	{1, "ext3", ext3_image},
 	{1, "ext2", ext2_image},
 	{1, "minix", minix_image},
 	{8, "reiserfs", reiserfs_image},
 	{64, "reiserfs", reiserfs_image},
+	{64, "reiser4", reiser4_image},
+	{64, "gfs2", gfs2_image},
 	{32, "jfs", jfs_image},
+	{32, "iso9660", iso_image},
 	{0, "luks", luks_image},
 	{0, "lvm2", lvm2_image},
 	{1, "lvm2", lvm2_image},
 	{-1, "swap", swap_image},
+	{-1, "suspend", suspend_image},
 	{0, "", NULL}
 };
 
